{-# LANGUAGE StandaloneDeriving, FlexibleInstances #-}

module Main ( main
            ) where

import Algos.Graphs.Forest(Forest(..), Tree(..), Subtree(..), preOrder, postOrder)
import Algos.Graphs.Graph(Graph(..), Vertex(..), Edge(..))
import Algos.Graphs.GraphIO(Orientation(..), Display(..), Parse(..), parseGraph, NewickType(..))
import Control.Applicative ((<$>))
import Data.Array (Array, array, accumArray, (!))
import Data.Monoid (Monoid(..))
import Data.List (find, tails, minimumBy, sortBy)
import Data.Map (Map)
import qualified Data.Map as Map (empty, lookupLT, lookupLE, insert)
import Data.Maybe (fromJust, listToMaybe, mapMaybe)
import Data.Ord (comparing)
import Data.Set (Set)
import qualified Data.Set as Set (fromList, member)
import System.Environment (getArgs, getProgName)
import System.Exit (exitFailure)

main :: IO ()
main =  do (gfile, ffile) <- readArgs
           (o, g) <- parseGraph  <$> readFile gfile
           f      <- parseNewick <$> readFile ffile
           putStrLn (formatResult $ verifyDFSForest o g f)

readArgs :: IO (String, String)
readArgs =  do args <- getArgs
               case args of
                 [gfile, ffile] -> return (gfile, ffile)
                 _              -> usage >> exitFailure

usage :: IO ()
usage =  do prog <- getProgName
            putStrLn $ "USAGE: " ++ prog ++ " <graph file> <forest file>"

formatResult                                      :: Result -> String
formatResult OK                                   =  "OK"
formatResult (MissingVertex v)                    =  "Vertex " ++ show v ++ " is missing from forest"
formatResult (NotAnEdge e)                        =  "Edge in forest is not in the graph: " ++ show e
formatResult (CrossEdge e)                        =  "Found a cross edge: " ++ show e
formatResult (ForwardCrossEdge e)                 =  "Found a forward cross edge: " ++ show e
formatResult (EarlyForwardEdge e@(_, u) f@(_, v)) =  "Edge " ++ show e ++ " succeeds edge " ++ show f ++ " but " ++
                                                     show u ++ " is a descendant of " ++ show v

data Result = OK
            | MissingVertex    Int
            | NotAnEdge        (Int, Int)
            | CrossEdge        (Int, Int)
            | ForwardCrossEdge (Int, Int)
            | EarlyForwardEdge (Int, Int) (Int, Int)

instance Monoid Result where
  mempty       = OK
  mappend OK x = x
  mappend x  _ = x

verifyDFSForest       :: Orientation -> Graph Int Int Int -> Forest Int Int -> Result
verifyDFSForest o g f =  mconcat [missingVertex, nonEdge, crossEdge, earlyForwardEdge]

  where fvs           = forestVertices f
        missingVertex = mconcat $ map (checkIfPresent fvs) (graphVertices g)
          
        ges     = graphEdgeSet g
        nonEdge = mconcat $ map (checkIfGraphEdge o ges) (forestEdges f)

        n         = length (graphVertices g)
        preNums   = array (1, n) (zip (preOrder  f) [1..])
        postNums  = array (1, n) (zip (postOrder f) [1..])
        crossEdge = mconcat $ map (checkForCrossEdge o preNums postNums) (graphEdges g)

        outNeighbours = accumArray (flip (:)) [] (1, n) (concatMap efun $ graphEdges g)
          where efun | o == Directed = \(Edge u v _) -> [(u, v)]
                     | otherwise     = \(Edge u v _) -> [(u, v), (v, u)]
        earlyForwardEdge = mconcat $ map (checkForEarlyForwardEdge preNums postNums outNeighbours) (graphVertices g)

forestVertices :: Forest Int Int -> Set Int
forestVertices =  Set.fromList . preOrder

forestEdges :: Forest Int Int -> [Edge Int Int]
forestEdges =  concatMap treeEdges . trees

treeEdges   :: Tree Int Int -> [Edge Int Int]
treeEdges t =  go t []
  where go (Tree _ []) es = es
        go (Tree v ws) es = map (\(Subtree e (Tree w _)) -> Edge v w e) ws ++
                            foldr go es (map (\(Subtree _ t) -> t) ws)

graphEdgeSet :: Graph Int Int Int -> Set (Edge Int Int)
graphEdgeSet =  Set.fromList . graphEdges

checkIfPresent                 :: Set Int -> Vertex Int Int -> Result
checkIfPresent vs (Vertex v _) |  v `Set.member` vs = OK
                               |  otherwise         = MissingVertex v

checkIfGraphEdge                     :: Orientation -> Set (Edge Int Int) -> Edge Int Int -> Result
checkIfGraphEdge o es e@(Edge u v l) |  e `Set.member` es                               = OK
                                     |  o == Undirected && (Edge v u l) `Set.member` es = OK
                                     |  otherwise                                       = NotAnEdge (u, v)

checkForCrossEdge                                 :: Orientation -> Array Int Int -> Array Int Int -> Edge Int Int -> Result
checkForCrossEdge o preNums postNums (Edge u v l) |  preU < preV && postU > postV = OK -- u is an ancestor of v
                                                  |  preV < preU && postV > postU = OK -- v is an ancestor of u
                                                  |  o == Directed && preV < preU = OK -- backward cross edge
                                                  |  o == Directed                = ForwardCrossEdge (u, v)
                                                  |  otherwise                    = CrossEdge (u, v)
  where  preU  = preNums  ! u
         preV  = preNums  ! v
         postU = postNums ! u
         postV = postNums ! v

checkForEarlyForwardEdge :: Array Int Int -> Array Int Int -> Array Int [Int] -> Vertex Int Int -> Result
checkForEarlyForwardEdge preNums postNums outNeighbours v = 
  case viol of
    Nothing     -> OK
    Just (u, c) -> EarlyForwardEdge (vertexId v, u) (vertexId v, c)
  where outnbs      = outNeighbours ! vertexId v
        descs       = zip (filter (\u -> preNums ! u > preNums ! vertexId v && postNums ! u < postNums ! vertexId v) 
                                  outnbs
                          ) [1..]
        chldrn      = foldl addChld Map.empty (sortBy (comparing $ (preNums !) . fst) descs)
        addChld m u = case Map.lookupLT numu m of
                        Nothing                                           -> m'
                        Just (_, c) | postNums ! fst c < postNums ! fst u -> m'
                                    | otherwise                           -> m
          where numu = preNums ! fst u
                m'   = Map.insert numu u m
        viol = listToMaybe $ mapMaybe chkViol descs
        chkViol (u, i) | j <= i    = Nothing
                       | otherwise = Just (u, c)
          where numu        = preNums ! u
                (_, (c, j)) = fromJust (Map.lookupLE numu chldrn)

instance Display Int where
  display = show

instance Parse Int where
  parse = read

deriving instance Eq (Vertex Int Int)
deriving instance Eq (Edge Int Int)

deriving instance Ord (Vertex Int Int)
deriving instance Ord (Edge Int Int)
