module Main ( main
            ) where

import Algos.Graphs.Graph (Graph(..), Vertex(..), Edge(..))
import Algos.Graphs.GraphIO (Display(..), Orientation(..), displayGraph)
import Algos.Random.Numbers (randomNumberM)
import Algos.Random.Partitions (randomSegmentedSampleM)
import Algos.Random.Sequences (randomPermutationM)
import Control.Applicative ((<$>), (<*>))
import Control.Exception (SomeException(..), handle)
import Control.Monad.State (State, evalState)
import Data.Array (listArray, (!))
import System.Environment (getArgs, getProgName)
import System.Exit (exitFailure)
import System.IO (withFile, IOMode(..), hPutStr)
import System.Random (RandomGen, newStdGen)

main :: IO ()
main =  do (n, m, o, file) <- readArgs
           gen             <- newStdGen
           let g = evalState (genGraph o n m) gen
           withFile file WriteMode $
             \h -> hPutStr h (displayGraph (o, g))

readArgs :: IO (Int, Int, Orientation, String)
readArgs =  handle (\(SomeException _) -> usage >> exitFailure) $
            do [nstr, mstr, ostr, file] <- getArgs
               let n = read nstr
                   m = read mstr
                   o = read ostr
               return (n, m, o, file)

usage :: IO ()
usage =  do prog <- getProgName
            putStrLn $ "USAGE: " ++ prog ++ " <n> <m> <Directed|Undirected> <output file>"

genGraph            :: RandomGen g => Orientation -> Int -> Int -> State g (Graph Int Int Int)
genGraph Directed   =  genGraph' dirMaxs   mkDirEdge
genGraph Undirected =  genGraph' undirMaxs mkUndirEdge

genGraph'                 :: RandomGen g => (Int -> [Int]) -> (Int -> Int -> State g (Int, Int)) ->
                             Int -> Int -> State g (Graph Int Int Int)
genGraph' maxs mkEdge n m =  do uss <- randomSegmentedSampleM m (maxs n)
                                eps <- sequence (concat $ 
                                                 zipWith (\us v -> zipWith mkEdge us (repeat v)) 
                                                         uss [1..n])
                                let m' = length eps
                                vids  <- listArray (1, n) <$> randomPermutationM [1..n]
                                elbls <- randomPermutationM [1..m']
                                let vs = [Vertex i 0 | i <- [1..n]]
                                    es = [Edge (vids!u) (vids!v) l | ((u, v), l) <- zip eps elbls]
                                Graph <$> randomPermutationM vs <*> randomPermutationM es

dirMaxs   :: Int -> [Int]
dirMaxs n =  take n (repeat (n-1))

mkDirEdge     :: Int -> Int -> State g (Int, Int)
mkDirEdge u v |  u < v     = return (u,   v)
              |  otherwise = return (u+1, v)

undirMaxs   :: Int -> [Int]
undirMaxs n =  [0..n-1]

mkUndirEdge     :: RandomGen g => Int -> Int -> State g (Int, Int)
mkUndirEdge u v =  do f <- randomNumberM (0,1)
                      case f of
                        0 -> return (u, v)
                        1 -> return (v, u)

instance Display Int where
  display = show
