module Algos.Random.Partitions ( randomSum
                               , randomSum'
                               , randomSumM
                               , randomSegmentedSum
                               , randomSegmentedSum'
                               , randomSegmentedSumM
                               , randomSegmentedSample
                               , randomSegmentedSample'
                               , randomSegmentedSampleM
                               , randomPartition
                               , randomPartition'
                               , randomPartitionM
                               , splitBySize
                               , splitByMax
                               ) where

import Algos.Random.Sequences (randomPermutationM)
import Algos.Random.Sets (randomCombinationM)
import Control.Applicative ((<$>), (<*>))
import Control.Monad.State (State, runState)
import Data.List (unfoldr)
import System.Random (RandomGen)

-- Computes a random partition of the number n into a sum of p numbers in
-- O(p lg p) time.  Each number is at least 1.
randomSum       :: RandomGen g => g -> Int -> Int -> [Int]
randomSum g p n =  fst $ randomSum' g p n

-- A version of randomSum that threads through the random number generator
randomSum'       :: RandomGen g => g -> Int -> Int -> ([Int], g)
randomSum' g p n =  runState (randomSumM p n) g

-- A State monad version of randomSum'
randomSumM     :: RandomGen g => Int -> Int -> State g [Int]
randomSumM p n =  do comb <- randomCombinationM (p-1) (n-1)
                     return (zipWith (-) (comb ++ [n]) (0:comb))

-- Given a list of p integers ns = [n1, n2, ...] and a value m, computes a list
-- of p numbers that sum to m and such that the ith number is between 0 and ni.
randomSegmentedSum        :: RandomGen g => g -> Int -> [Int] -> [Int]
randomSegmentedSum g m ns =  fst $ randomSegmentedSum' g m ns

-- A version of randomSegmentedSum that threads through the random number
-- generator
randomSegmentedSum'        :: RandomGen g => g -> Int -> [Int] -> ([Int], g)
randomSegmentedSum' g m ns =  runState (randomSegmentedSumM m ns) g

-- A State monad version of randomSegmentedSum'
randomSegmentedSumM      :: RandomGen g => Int -> [Int] -> State g [Int]
randomSegmentedSumM m ns =  map length <$> randomSegmentedSampleM m ns

-- Given a list of integers ns = [n1, n2, ...], computes a list of lists such
-- that the ith list contains numbers between 1 and ni and in total p numbers
-- are sampled.  Any such segmented sample is equally likely to be produced.
randomSegmentedSample        :: RandomGen g => g -> Int -> [Int] -> [[Int]]
randomSegmentedSample g p ns =  fst $ randomSegmentedSample' g p ns

-- A version of randomSegmentedSample that threads through the random number
-- generator
randomSegmentedSample'        :: RandomGen g => g -> Int -> [Int] -> ([[Int]], g)
randomSegmentedSample' g p ns =  runState (randomSegmentedSampleM p ns) g

-- A State monad version of randomSegmentedSample'
randomSegmentedSampleM      :: RandomGen g => Int -> [Int] -> State g [[Int]]
randomSegmentedSampleM p ns =  zipWith adjustSample accs .
                               splitByMax (tail accs) <$> 
                               randomCombinationM p (sum ns)
  where accs                = scanl (+) 0 ns
        adjustSample acc ss = map (subtract acc) ss

-- Computes a random partition of a sequence of into p subsequences in 
-- O(n + p lg p) time.
randomPartition        :: RandomGen g => g -> Int -> [a] -> [[a]]
randomPartition g p xs =  fst $ randomPartition' g p xs

-- A version of randomPartition that threads through the random number generator
randomPartition'        :: RandomGen g => g -> Int -> [a] -> ([[a]], g)
randomPartition' g p xs =  runState (randomPartitionM p xs) g

-- A State monad version of randomPartition'
randomPartitionM      :: RandomGen g => Int -> [a] -> State g [[a]]
randomPartitionM p xs =  splitBySize <$> randomSumM p (length xs)
                                     <*> randomPermutationM xs

-- Splits a given list into sublists of the lengths specified in the second list
splitBySize :: [Int] -> [a] -> [[a]]  
splitBySize =  curry (unfoldr doSplit)
  where doSplit ([],   _ ) = Nothing
        doSplit (n:ns, xs) = Just (xs', (ns, xs''))
          where (xs', xs'') = splitAt n xs

-- Splits an increasing list into sublists based on an increasing list of
-- upper bounds on the values in each sublist.  For example, splitByMax
-- [1,5,7,11] [2,6,8,10] = [[], [], [6], [8, 10]]
splitByMax :: Ord a => [a] -> [a] -> [[a]]
splitByMax =  curry (unfoldr doSplit)
  where doSplit ([],       _ ) = Nothing
        doSplit (mx : mxs, xs) = Just (xs', (mxs, xs''))
          where (xs', xs'') = span (<= mx) xs
