{-# LANGUAGE TupleSections #-}

module Algos.Random.Sequences ( randomPermutation
                              , randomPermutation'
                              , randomPermutationM
                              , randomSubPermutation
                              , randomSubPermutation'
                              , randomSubPermutationM
                              ) where

import Algos.Random.Impl (floydM)
import Control.Applicative ((<$>))
import Control.Monad.State (State, state, runState)
import Control.Monad.ST (ST, runST)
import Data.Array.ST (STArray, getElems, readArray, writeArray, newListArray)
import Data.Map (member, insert, adjust)
import System.Random (RandomGen, randomR)

-- Computes a uniform random permutation of the given sequence in linear time
randomPermutation      :: RandomGen g => g -> [a] -> [a]
randomPermutation g xs =  fst $ randomPermutation' g xs

-- A version of randomPermutation that threads through the random number generator
randomPermutation'      :: RandomGen g => g -> [a] -> ([a], g)
randomPermutation' g [] =  ([], g)
randomPermutation' g xs =  runST (newListArray (1, n) xs >>= permuteA g n)
  where n = length xs

-- The core of randomPermutation'.  Uses an STArray to achieve linear time.
permuteA        :: RandomGen g => g -> Int -> STArray s Int a -> ST s ([a], g)
permuteA g n xs =  perm g n
  where  perm g 1 = (,g) <$> getElems xs
         perm g i = do let (j, g') = randomR (1, i) g
                       x_i  <- readArray xs i
                       x_j  <- readArray xs j
                       writeArray xs j x_i
                       writeArray xs i x_j
                       perm g' (i-1)

-- A State monad version of randomPermutation'
randomPermutationM    :: RandomGen g => [a] -> State g [a]
randomPermutationM xs =  state (flip randomPermutation' xs)

-- Computes a random permutation of p integers betewen 1 and n in O(p lg p) time.
randomSubPermutation       :: RandomGen g => g -> Int -> Int -> [Int]
randomSubPermutation g p n =  fst $ randomSubPermutation' g p n

-- A version of randomSubPermutation that threads through the random number generator
randomSubPermutation'       :: RandomGen g => g -> Int -> Int -> ([Int], g)
randomSubPermutation' g p n =  runState (randomSubPermutationM p n) g

-- A State monad version of randomSubPermutation
randomSubPermutationM :: RandomGen g => Int -> Int -> State g [Int]
randomSubPermutationM =  floydM addNum toSeq
  where addNum (ts, m) (t, j) | member t m = (ts,   insert j [] $ adjust (j:) t m)
                              | otherwise  = (t:ts, insert t [] m)
        toSeq                              = id
